from OFS import SimpleItem
from OFS.PropertyManager import PropertyManager
import Globals # so I can get Globals data_dir
from Globals import HTMLFile
from DateTime import DateTime

import os, os.path, cPickle, struct, operator

"""
There can be only one! Notice that only one instance if this product 
can be installed on a zope server!!!! Trying to installe more than one 
instance will silently fail. The counter will still work, of it is in 
the aquisition path.

Probably this should be made placeless. Well that is for another 
version.
"""


#################
# it does not save state between restarts of the server!!!!!!!!!!
# I will handle this later


# conflic handlers
from threading import Lock
l = Lock()

# where the counters dict is saved to
file_name = str(Globals.data_dir) + os.sep + 'counters.dict'
unsaved_counts = 0

def file_exists():
    "Returns true if file allready exists"
    return os.path.exists(file_name)

def get_from_file():
    "gets the counters dict from the file, or a new empty dict"
    l.acquire()
    try:
        if file_exists():
            f = open(file_name, 'rb')
            c = cPickle.load(f)
            f.close()
            return c
        else:
            return {}
    finally:
        l.release()

# module variable for shared volatile state between threads
counters = get_from_file()

def save_to_file():
    "Saves the countes dict to the file"
    try:
        f = open(file_name, 'wb')
        cPickle.dump(counters, f, -1) 
        f.close()
    except:
        pass


def delete_file():
    "Deletes the counter file. Only used when counter is uninstalled."
    try:
        os.remove(file_name)
    except:
        l.release()


def increase_count(key, save_interval):
    global unsaved_counts
    l.acquire()
    try:
        new_count = counters[key] = counters.get(key, 0) + 1
        counters[key] = new_count
        unsaved_counts += 1
        if unsaved_counts >= save_interval:
            save_to_file()
            unsaved_counts = 0
        return new_count
    finally:
        l.release()


def get_count(key):
    l.acquire()
    try:
        return counters.get(key, 0)
    finally:
        l.release()


def set_unsaved_counts(value):
    global unsaved_counts
    l.acquire()
    try:
        unsaved_counts = value
    finally:
        l.release()

def get_all():
    "Returns the counters dictionary"
    l.acquire()
    try:
        return counters
    finally:
        l.release()


def del_counts(paths):
    "Delete the counters from the counters dict"
    l.acquire()
    try:
        for path in paths:
            if counters.has_key(path):
                del(counters[path])
    finally:
        l.release()


def set_counts(paths, new_value):
    "Delete the counters from the counters dict"
    l.acquire()
    try:
        for path in paths:
            counters[path] = new_value
    finally:
        l.release()


class mxmCounter(PropertyManager, SimpleItem.SimpleItem):

    "mxmCounter, simple hit count and statistics"

    meta_type = 'mxmCounter'
    
    id = 'mxm_counter'

    _properties = (
        {'id':'title', 'type':'string', 'mode':'w'},
        {'id':'ips_to_ignore', 'type':'lines', 'mode':'w'},
        {'id':'count_once', 'type':'boolean', 'mode':'w'}, # use cookie
        {'id':'save_interval', 'type':'int', 'mode':'w'},
    )


    manage_options = PropertyManager.manage_options + SimpleItem.SimpleItem.manage_options + (
    	{'label': 'Set counters', 'action': 'manage_setCounters',},
    )


    __ac_permissions__ = PropertyManager.__ac_permissions__ +\
                         SimpleItem.SimpleItem.__ac_permissions__ + (
        # label
#        ('View',
#            # methods
#            ('index_html',),
#            # roles
#            ('Anonymous', 'Manager'),
#        ),
        ('Change mxmCounter',
            ('manage_edit_mxmCounter', 'manage_editForm_mxmCounter', 
            'manage_setCounters', 'manage_setCountersAction',),
            ('Manager',)
        ),
        ('Add mxmCounter',
            ('manage_addAction_mxmCounter', 'manage_addForm_mxmCounter'),
            ('Manager',)
        ),
    )


    def __init__(self, title='', ips_to_ignore=None, count_once=0, save_interval=100):
        "Inits the product with default values"
        if ips_to_ignore is None:
            ips_to_ignore = []
        self._edit(title=title, ips_to_ignore=ips_to_ignore, 
                   count_once=count_once, save_interval=save_interval)


    def _edit(self, **args):
        "internal helper"
        for key, val in args.items():
            setattr(self, key, val)


    def _ignore_ip(self, ip):
        "returns true if ip adress should be ignored"
        ips_to_ignore = [i.strip() for i in self.ips_to_ignore if i.strip()]
        for ip_to_ignore in ips_to_ignore:
            if ip.startswith(ip_to_ignore):
                return 1
        return 0
        
    
    def count(self, obj, REQUEST=None):
        """
        Increases The Count For The Object. It Uses The Physical Path 
        as a key to the count. So it will loose count if object is 
        moved. Tough luck!
        """
        req = obj.REQUEST
        url_path = req.URLPATH0
        # try to create unique cookie name
        # I will change it if somebody bitches
        # cookie lasts for a session (pr. browser restart)
        cookie_name = self.meta_type + url_path
        cookie_is_set = req.has_key(cookie_name)
        # if ip is to be ignored, we should only return the value, not 
        # increase it
        ip = req.get('HTTP_X_FORWARDED_FOR', '') or req.get('REMOTE_ADDR', '')
        if (ip and self._ignore_ip(ip)) or (self.count_once and cookie_is_set):
            return self.get_count(obj, req)
        else:
            if self.count_once:
                req.response.setCookie(cookie_name, 1)
            return increase_count(url_path, self.save_interval)
        
        
    def get_count(self, obj, REQUEST=None):
        "returns the count without incrementing"
        url_path = obj.REQUEST.URLPATH0
        return get_count(url_path)
        

    def getSortedByHits(self):
        "Returns the [(url_path, hit)] sorted by hits"
        c = get_all()
        decorated = [(val, key) for (key, val) in c.items()]
        decorated.sort()
        decorated.reverse()
        return [(val, key) for (key, val) in decorated]


    def total_hits(self):
        "Returns the total number of hits on all counted pages"
        c = get_all()
        j = [val for (key, val) in c.items()]
        return reduce(operator.add, j, 0)

    def manage_beforeDelete(self, item, container):
        "Remember to delete the counters file before letting the object go"
        delete_file()
        SimpleItem.SimpleItem.manage_beforeDelete(self, item, container)
        

    def manage_setCountersAction(self, counter_paths=None, new_value=0, REQUEST=None):
        "Set the new value for selcted counters"
        if new_value == 0:
            del_counts(counter_paths)
        else:
            set_counts(counter_paths, new_value)
        save_to_file()
        if REQUEST is not None:
            return self.manage_setCounters(self, REQUEST)
    
    
    # renders mxmCounter statistics
    manage_setCounters = HTMLFile('www/manage_setCounters', globals())


#####################################################
# Constructor functions, only used when adding class
# to objectManager


def manage_addAction_mxmCounter(self, REQUEST=None):
    "Add an arbitrary class to a folder to a folder."
    # if no id passed, generate one from the datetime
    # we silently fail if instance allready exists
    allready_exists = file_exists()
    if not allready_exists:
        id = mxmCounter.id
        self._setObject(id, mxmCounter())
        save_to_file() # save initial empty counter dict to file
    if REQUEST is not None:
        if not allready_exists:
            getattr(self, id).manage_changeProperties(REQUEST)
        return self.manage_main(self, REQUEST)


#manage_addForm_mxmCounter = HTMLFile('manage_addForm_mxmCounter', globals())


constructors = (manage_addAction_mxmCounter, ) #manage_addForm_mxmCounter, 

